@@ -35,7 +35,8 @@
 	 map_field_lists/1,cover_bin_opt/1,
 	 val_dsetel/1,bad_tuples/1,bad_try_catch_nesting/1,
          receive_stacked/1,aliased_types/1,type_conflict/1,
-         infer_on_eq/1,infer_dead_value/1]).
+         infer_on_eq/1,infer_dead_value/1,
+         receive_marker/1,safe_instructions/1]).
 
 -include_lib("common_test/include/ct.hrl").
 
@@ -65,7 +66,7 @@ groups() ->
        map_field_lists,cover_bin_opt,val_dsetel,
        bad_tuples,bad_try_catch_nesting,
        receive_stacked,aliased_types,type_conflict,
-       infer_on_eq,infer_dead_value]}].
+       infer_on_eq,infer_dead_value,receive_marker]}].
 
 init_per_suite(Config) ->
     test_lib:recompile(?MODULE),
@@ -681,11 +682,16 @@ infer_on_eq_4(T) ->
 
 %% ERIERL-348; types were inferred for dead values, causing validation to fail.
 
+-record(idv, {key}).
+
 infer_dead_value(Config) when is_list(Config) ->
     a = idv_1({a, b, c, d, e, f, g}, {0, 0, 0, 0, 0, 0, 0}),
     b = idv_1({a, b, c, d, 0, 0, 0}, {a, b, c, d, 0, 0, 0}),
     c = idv_1({0, 0, 0, 0, 0, f, g}, {0, 0, 0, 0, 0, f, g}),
     error = idv_1(gurka, gaffel),
+
+    ok = idv_2(id(#idv{})),
+
     ok.
 
 idv_1({_A, _B, _C, _D, _E, _F, _G},
@@ -700,6 +706,44 @@ idv_1({_A, _B, _C, _D, _E, F, G},
 idv_1(_A, _B) ->
     error.
 
+%% ERL-995: The first solution to ERIERL-348 was incomplete and caused
+%% validation to fail when living values depended on delayed type inference on
+%% "dead" values.
+
+idv_2(State) ->
+    Flag = (State#idv.key == undefined),
+    case id(gurka) of
+        {_} -> id([Flag]);
+        _ -> ok
+    end,
+    if
+        Flag -> idv_called_once(State);
+        true -> ok
+    end.
+
+idv_called_once(_State) -> ok.
+
+receive_marker(Config) when is_list(Config) ->
+    Errors = do_val(receive_marker, Config),
+
+    [{{receive_marker,t1,1},
+      {return,_,
+       {return_with_receive_marker,committed}}},
+     {{receive_marker,t2,1},
+      {{call_last,1,{f,2},1},_,
+       {return_with_receive_marker,committed}}}] = Errors,
+
+    ok.
+
+%% ERL-1128: the validator erroneously thought that many non-throwing
+%% instructions like is_eq_exact could throw.
+safe_instructions(Config) when is_list(Config) ->
+    Errors = do_val(safe_instructions, Config),
+
+    [] = Errors,
+
+    ok.
+
 %%%-------------------------------------------------------------------------
 
 transform_remove(Remove, Module) ->
